package de.hft.stuttgart.citydoctor2.checks.geometry;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.checks.util.CollectionUtils;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;

/**
 * This class checks if a solid geometry has polygons that are fully enclosed inside the boundary surfaces.
 * 
 * @author Alexander Riegel
 */
public class InteriorFacesCheck extends Check {

	private static final List<CheckId> dependencies;

	static {
		ArrayList<CheckId> deps = new ArrayList<>();
		deps.add(CheckId.C_GE_R_TOO_FEW_POINTS);
		deps.add(CheckId.C_GE_R_NOT_CLOSED);
		deps.add(CheckId.C_GE_R_DUPLICATE_POINT);
		deps.add(CheckId.C_GE_R_SELF_INTERSECTION);
		deps.add(CheckId.C_GE_P_INTERIOR_DISCONNECTED);
		deps.add(CheckId.C_GE_P_INTERSECTING_RINGS);
		deps.add(CheckId.C_GE_P_NON_PLANAR);
		deps.add(CheckId.C_GE_S_TOO_FEW_POLYGONS);
		dependencies = Collections.unmodifiableList(deps);
	}

	@Override
	public void check(Geometry g) {
		
		// only check solids
		if (g.getType() != GeometryType.SOLID) {
			return;
		}
		// TODO: Implement check logic
		// Assuming geometry is well defined (no holes or dangling polygons).
		// Check should run just before the ManifoldEdgesCheck
		// Step 1:
		// Check pairs of polygons for their distance
		// if they are too closely together (within EPSILON) add them to confirmed set
		
		// Step 2:
		// get centroid of each polygon and move it EPSILON * normalized polygon normal vector
		// if that point is inside the solid's geometry, add polygon to confirmed set
		
		// Step 3:
		// Use intersections of the polygon normal vector(raycast from centroid) 
		// with other polygons to establish a list of suspects.
		// Check all found suspects for number of intersections.
		// Use intersection count to eliminate wrongly suspected polygons
		// After sorting in ascending number of intersections:
		// 0 intersections are safe to remove from suspects (safe)
		// 1 intersection is guaranteed to be an interior face (confirmed)
		// Move these polygons into their respective sets
		
		// Intersection counts >= 2 require more analysis,
		// using dot product of the normalized polygon normal vectors.
		// Count the number of positive (same face direction) and negative (opposite face directions) dot products.
		// If the intersected polygon is in the confirmed set, ignore them
		// If positive count = negative count, move into preliminary safe set
		//
		
	}

	@Override
	public Set<Requirement> appliesToRequirements() {
		return CollectionUtils.singletonSet(Requirement.R_GE_S_INTERIOR_FACE);
	}

	@Override
	public List<CheckId> getDependencies() {
		return dependencies;
	}

	@Override
	public RequirementType getType() {
		return RequirementType.GEOMETRY;
	}

	@Override
	public Check createNewInstance() {
		return new InteriorFacesCheck();
	}

	@Override
	public CheckId getCheckId() {
		return CheckId.C_GE_S_INTERIOR_FACE;
	}

}
